/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package rest

import (
	"bytes"
	"context"
	"io/ioutil"
	"net/http"
	"net/http/httptest"
	"os"
	"testing"

	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/crt"
	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/pem"

	"github.com/sirupsen/logrus"

	api_v1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes/fake"
	core_v1_listers "k8s.io/client-go/listers/core/v1"
	"k8s.io/client-go/tools/cache"
)

func TestHealthz(t *testing.T) {
	version := "47.11"
	hndlr := &healthzHndlr{
		version: version,
		log:     &logrus.Logger{Out: ioutil.Discard},
	}
	rr := httptest.NewRecorder()

	req := httptest.NewRequest(http.MethodGet, "/v1/healthz", nil)
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNoContent {
		t.Errorf("GET /v1/healthz status: got %d want %d",
			rr.Code, http.StatusNoContent)
	}
	if rr.Result().Header.Get("Application-Version") != version {
		t.Errorf("GET /v1/healthz Application-Version: got %s want %s",
			rr.Result().Header.Get("Application-Version"),
			version)
	}
	textPlain := "text/plain"
	if rr.Result().Header.Get("Content-Type") != textPlain {
		t.Errorf("GET /v1/healthz Content-Type: got %s want %s",
			rr.Result().Header.Get("Content-Type"),
			textPlain)
	}
	noStore := "no-store"
	if rr.Result().Header.Get("Cache-Control") != noStore {
		t.Errorf("GET /v1/healthz Cache-Control: got %s want %s",
			rr.Result().Header.Get("Cache-Control"),
			noStore)
	}
	if rr.Result().Header.Get("Content-Length") != "" {
		t.Errorf("GET /v1/healthz Content-Length: got %s want empty",
			rr.Result().Header.Get("Content-Length"))
	}

	req = httptest.NewRequest(http.MethodHead, "/v1/healthz", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNoContent {
		t.Errorf("HEAD /v1/healthz status: got %d want %d",
			rr.Code, http.StatusNoContent)
	}

	req = httptest.NewRequest(http.MethodGet, "/v1/healthz/foo", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("GET /v1/healthz/foo status: got %d want %d",
			rr.Code, http.StatusNotFound)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("GET /v1/healthz/foo Content-Length: got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}
}

func TestHealthz405(t *testing.T) {
	version := "08.15"
	hndlr := &healthzHndlr{
		version: version,
		log:     &logrus.Logger{Out: ioutil.Discard},
	}
	textPlain := "text/plain"
	noStore := "no-store"

	methods := []string{
		http.MethodPost, http.MethodPut, http.MethodPatch,
		http.MethodDelete, http.MethodConnect, http.MethodOptions,
		http.MethodTrace, "FOOBAR",
	}
	for _, method := range methods {
		req := httptest.NewRequest(method, "/v1/healthz", nil)
		rr := httptest.NewRecorder()
		hndlr.ServeHTTP(rr, req)
		if rr.Code != http.StatusMethodNotAllowed {
			t.Errorf("%s /v1/healthz status: got %d want %d",
				method, rr.Code, http.StatusMethodNotAllowed)
		}
		if rr.Result().Header.Get("Content-Length") != "0" {
			t.Errorf("%s /v1/healthz Content-Length: got %s want 0",
				method, rr.Result().Header.
					Get("Content-Length"))
		}
		if rr.Result().Header.Get("Content-Type") != textPlain {
			t.Errorf("%s /v1/healthz Content-Type: got %s want %s",
				method, rr.Result().Header.Get("Content-Type"),
				textPlain)
		}
		if rr.Result().Header.Get("Cache-Control") != noStore {
			t.Errorf("%s /v1/healthz Cache-Control: got %s want %s",
				method, rr.Result().Header.Get("Cache-Control"),
				noStore)
		}
	}
}

func setupSecretLister(client *fake.Clientset) core_v1_listers.SecretLister {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	infFactory := informers.NewSharedInformerFactory(client, 0)
	informer := infFactory.Core().V1().Secrets().Informer()
	lister := infFactory.Core().V1().Secrets().Lister()

	infFactory.Start(ctx.Done())
	cache.WaitForCacheSync(ctx.Done(), informer.HasSynced)
	return lister
}

func TestPutPem(t *testing.T) {
	gid := os.Getgid()
	base, err := ioutil.TempDir("", "k8s-crt-dnldr_rest_test")
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns1",
				Name:            "secret1",
				UID:             "2572e346-0710-4c3d-a1ae-bacaddb6f3f9",
				ResourceVersion: "4711",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crt1"),
				"tls.key": []byte("key1"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns2",
				Name:            "secret2",
				UID:             "1ce39371-b2a4-43f7-9e1d-240db0b933a4",
				ResourceVersion: "0815",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crt2"),
				"tls.key": []byte("key2"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns3",
				Name:            "secret3",
				UID:             "204ebad6-e06d-4e94-a0c5-82d5968b4bc4",
				ResourceVersion: "123456",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crt3"),
				"tls.key": []byte("key3"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns4",
				Name:            "secret4",
				UID:             "97b45482-bf35-4949-9319-fb2f9aa1eeac",
				ResourceVersion: "654321",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crt4"),
				"tls.key": []byte("key4"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "missing",
				Name:            "crtField",
				UID:             "ffefdcf1-0a54-4545-9918-ffd67f754360",
				ResourceVersion: "1",
			},
			Data: map[string][]byte{
				"foo":     []byte("fooVal"),
				"tls.key": []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "missing",
				Name:            "keyField",
				UID:             "37c1d44d-eb10-4b0b-8398-b35e10841bf0",
				ResourceVersion: "2",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crtSecret"),
				"foo":     []byte("fooVal"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "empty",
				Name:            "crtField",
				UID:             "d63ef37e-2ffe-44b7-9e3e-ba84d96e8d95",
				ResourceVersion: "398428",
			},
			Data: map[string][]byte{
				"tls.crt": []byte{},
				"tls.key": []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "empty",
				Name:            "keyField",
				UID:             "1a5716f0-b996-4295-8565-1d1a7037edcf",
				ResourceVersion: "229880",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crtSecret"),
				"tls.key": []byte{},
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "invalid",
				Name:            "typeField",
				UID:             "c7e6edc6-f8c2-4845-810a-938f9d6a5d5f",
				ResourceVersion: "3",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crtSecret"),
				"tls.key": []byte("keySecret"),
			},
			Type: api_v1.SecretType("foo/bar"),
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	files := pem.NewFiles(base, gid, getter)

	hndlr := &pemsHndlr{
		log:       &logrus.Logger{Out: ioutil.Discard},
		files:     files,
		crtGetter: getter,
	}

	req := httptest.NewRequest(http.MethodPut, "/v1/pems/ns1/secret1", nil)
	rr := httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT /v1/pems/ns1/secret1 status: got %d want %d",
			rr.Code, http.StatusCreated)
	}
	if !files.Check("ns1", "secret1",
		"2572e346-0710-4c3d-a1ae-bacaddb6f3f9", "4711") {
		t.Error("files.Check() after PUT: got false want true")
	}
	file, ok := files.Files["ns1/secret1"]
	if !ok {
		t.Fatal("no file map entry after PUT /v1/pems/ns1/secret1")
	}
	path := files.Path(file)
	expected := []byte("crt1\nkey1")
	if contents, err := ioutil.ReadFile(path); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", path, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}
	textPlain := "text/plain"
	if rr.Result().Header.Get("Content-Type") != textPlain {
		t.Errorf("PUT /v1/pems/ns1/secret1 Content-Type: "+
			"got %s want %s",
			rr.Result().Header.Get("Content-Type"),
			textPlain)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("PUT /v1/pems/ns1/secret1 Content-Length: "+
			"got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}

	req = httptest.NewRequest(http.MethodPut, "/v1/pems/ns1/secret1", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNoContent {
		t.Errorf("2nd PUT /v1/pems/ns1/secret1 status: got %d want %d",
			rr.Code, http.StatusNoContent)
	}
	if rr.Result().Header.Get("Content-Type") != textPlain {
		t.Errorf("PUT /v1/pems/ns1/secret1 Content-Type: "+
			"got %s want %s",
			rr.Result().Header.Get("Content-Type"),
			textPlain)
	}
	if rr.Result().Header.Get("Content-Length") != "" {
		t.Errorf("2nd PUT /v1/pems/ns1/secret1 Content-Length: "+
			"got %s want empty",
			rr.Result().Header.Get("Content-Length"))
	}

	req = httptest.NewRequest(http.MethodPut, "/v1/pems/foo/bar", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("PUT /v1/pems/foo/bar status: got %d want %d",
			rr.Code, http.StatusNotFound)
	}
	if rr.Result().Header.Get("Content-Type") != textPlain {
		t.Errorf("PUT /v1/pems/foo/bar Content-Type: "+
			"got %s want %s",
			rr.Result().Header.Get("Content-Type"),
			textPlain)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("PUT /v1/pems/foo/bar Content-Length: "+
			"got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}

	req = httptest.NewRequest(http.MethodPut,
		"/v1/pems/ns2/secret2?uid=1ce39371-b2a4-43f7-9e1d-240db0b933a4"+
			"&version=0815", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT /v1/pems/ns2/secret2 status: got %d want %d",
			rr.Code, http.StatusCreated)
	}
	if !files.Check("ns2", "secret2",
		"1ce39371-b2a4-43f7-9e1d-240db0b933a4", "0815") {
		t.Error("files.Check() after PUT: got false want true")
	}
	file, ok = files.Files["ns2/secret2"]
	if !ok {
		t.Fatal("no file map entry after PUT /v1/pems/ns2/secret2")
	}
	path = files.Path(file)
	expected = []byte("crt2\nkey2")
	if contents, err := ioutil.ReadFile(path); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", path, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}

	req = httptest.NewRequest(http.MethodPut,
		"/v1/pems/ns3/secret3?uid=204ebad6-e06d-4e94-a0c5-82d5968b4bc4",
		nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT /v1/pems/ns3/secret3 status: got %d want %d",
			rr.Code, http.StatusCreated)
	}
	if !files.Check("ns3", "secret3",
		"204ebad6-e06d-4e94-a0c5-82d5968b4bc4", "123456") {
		t.Error("files.Check() after PUT: got false want true")
	}
	file, ok = files.Files["ns3/secret3"]
	if !ok {
		t.Fatal("no file map entry after PUT /v1/pems/ns3/secret3")
	}
	path = files.Path(file)
	expected = []byte("crt3\nkey3")
	if contents, err := ioutil.ReadFile(path); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", path, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}

	req = httptest.NewRequest(http.MethodPut,
		"/v1/pems/ns4/secret4?version=654321", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT /v1/pems/ns4/secret4 status: got %d want %d",
			rr.Code, http.StatusCreated)
	}
	if !files.Check("ns4", "secret4",
		"97b45482-bf35-4949-9319-fb2f9aa1eeac", "654321") {
		t.Error("files.Check() after PUT: got false want true")
	}
	file, ok = files.Files["ns4/secret4"]
	if !ok {
		t.Fatal("no file map entry after PUT /v1/pems/ns4/secret4")
	}
	path = files.Path(file)
	expected = []byte("crt4\nkey4")
	if contents, err := ioutil.ReadFile(path); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", path, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}

	req = httptest.NewRequest(http.MethodPut, "/v1/pems/missing/crtField",
		nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusForbidden {
		t.Errorf("PUT /v1/pems/missing/crtField status: got %d want %d",
			rr.Code, http.StatusForbidden)
	}
	// XXX check response body
	if files.Check("missing", "crtField", "", "") {
		t.Error("files.Check() after PUT: got true want false")
	}

	req = httptest.NewRequest(http.MethodPut, "/v1/pems/missing/keyField",
		nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusForbidden {
		t.Errorf("PUT /v1/pems/missing/keyField status: got %d want %d",
			rr.Code, http.StatusForbidden)
	}
	// XXX check response body
	if files.Check("missing", "keyField", "", "") {
		t.Error("files.Check() after PUT: got true want false")
	}

	req = httptest.NewRequest(http.MethodPut, "/v1/pems/empty/crtField",
		nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusForbidden {
		t.Errorf("PUT /v1/pems/empty/crtField status: got %d want %d",
			rr.Code, http.StatusForbidden)
	}
	// XXX check response body
	if files.Check("empty", "crtField", "", "") {
		t.Error("files.Check() after PUT: got true want false")
	}

	req = httptest.NewRequest(http.MethodPut, "/v1/pems/empty/keyField",
		nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusForbidden {
		t.Errorf("PUT /v1/pems/empty/keyField status: got %d want %d",
			rr.Code, http.StatusForbidden)
	}
	// XXX check response body
	if files.Check("empty", "keyField", "", "") {
		t.Error("files.Check() after PUT: got true want false")
	}

	req = httptest.NewRequest(http.MethodPut, "/v1/pems/invalid/typeField",
		nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusForbidden {
		t.Errorf("PUT /v1/pems/invalid/typeField status: "+
			"got %d want %d", rr.Code, http.StatusForbidden)
	}
	// XXX check response body
	if files.Check("invalid", "typeField", "", "") {
		t.Error("files.Check() after PUT: got true want false")
	}
}

func TestPostPem(t *testing.T) {
	gid := os.Getgid()
	base, err := ioutil.TempDir("", "k8s-crt-dnldr_rest_test")
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns1",
				Name:            "secret1",
				UID:             "eeb81cb6-7ec8-4095-bb15-933bbe528f70",
				ResourceVersion: "408198",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crt1"),
				"tls.key": []byte("key1"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	files := pem.NewFiles(base, gid, getter)

	hndlr := &pemsHndlr{
		log:       &logrus.Logger{Out: ioutil.Discard},
		files:     files,
		crtGetter: getter,
	}

	req := httptest.NewRequest(http.MethodPost, "/v1/pems/ns1/secret1", nil)
	rr := httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("POST /v1/pems/ns1/secret1 status: got %d want %d",
			rr.Code, http.StatusCreated)
	}
	if !files.Check("ns1", "secret1",
		"eeb81cb6-7ec8-4095-bb15-933bbe528f70", "408198") {
		t.Error("files.Check() after POST: got false want true")
	}
	file, ok := files.Files["ns1/secret1"]
	if !ok {
		t.Fatal("no file map entry after POST /v1/pems/ns1/secret1")
	}
	path := files.Path(file)
	expected := []byte("crt1\nkey1")
	if contents, err := ioutil.ReadFile(path); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after POST: %+v", path, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after POST: got %q want %q", contents,
			expected)
	}
	textPlain := "text/plain"
	if rr.Result().Header.Get("Content-Type") != textPlain {
		t.Errorf("POST /v1/pems/ns1/secret1 Content-Type: "+
			"got %s want %s",
			rr.Result().Header.Get("Content-Type"),
			textPlain)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("POST /v1/pems/ns1/secret1 Content-Length: "+
			"got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}

	req = httptest.NewRequest(http.MethodPost, "/v1/pems/ns1/secret1", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusConflict {
		t.Errorf("2nd POST /v1/pems/ns1/secret1 status: got %d want %d",
			rr.Code, http.StatusConflict)
	}
	if rr.Result().Header.Get("Content-Type") != textPlain {
		t.Errorf("POST /v1/pems/ns1/secret1 Content-Type: "+
			"got %s want %s",
			rr.Result().Header.Get("Content-Type"),
			textPlain)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("2nd POST /v1/pems/ns1/secret1 Content-Length: "+
			"got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}
}

func TestDeletePem(t *testing.T) {
	gid := os.Getgid()
	base, err := ioutil.TempDir("", "k8s-crt-dnldr_rest_test")
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns1",
				Name:            "secret1",
				UID:             "5b91fc9e-ef27-4edd-944c-44e1eb4714a3",
				ResourceVersion: "252729",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crt1"),
				"tls.key": []byte("key1"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns2",
				Name:            "secret2",
				UID:             "ef2022d5-8f11-4a27-bccb-2f17e1433bf1",
				ResourceVersion: "392440",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crt2"),
				"tls.key": []byte("key2"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	files := pem.NewFiles(base, gid, getter)

	hndlr := &pemsHndlr{
		log:       &logrus.Logger{Out: ioutil.Discard},
		files:     files,
		crtGetter: getter,
	}

	if found, valid, err := files.Write("ns1", "secret1", "", ""); !found ||
		!valid || err != nil {
		t.Fatalf("Write() failed: found=%v valid=%v err=%v", found,
			valid, err)
	}
	req := httptest.NewRequest(http.MethodDelete, "/v1/pems/ns1/secret1",
		nil)
	rr := httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNoContent {
		t.Errorf("DELETE /v1/pems/ns1/secret1 status: got %d want %d",
			rr.Code, http.StatusNoContent)
	}
	if files.Check("ns1", "secret1", "", "") {
		t.Error("files.Check() after DELETE: got true want false")
	}
	textPlain := "text/plain"
	if rr.Result().Header.Get("Content-Type") != textPlain {
		t.Errorf("DELETE /v1/pems/ns1/secret1 Content-Type: "+
			"got %s want %s",
			rr.Result().Header.Get("Content-Type"),
			textPlain)
	}
	if rr.Result().Header.Get("Content-Length") != "" {
		t.Errorf("DELETE /v1/pems/ns1/secret1 Content-Length: "+
			"got %s want empty",
			rr.Result().Header.Get("Content-Length"))
	}

	req = httptest.NewRequest(http.MethodDelete, "/v1/pems/ns1/secret1",
		nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("2nd DELETE /v1/pems/ns1/secret1 status: "+
			"got %d want %d", rr.Code, http.StatusNotFound)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("DELETE /v1/pems/ns1/secret1 Content-Length: "+
			"got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}

	req = httptest.NewRequest(http.MethodDelete, "/v1/pems/foo/bar", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("DELETE /v1/pems/foo/bar status: got %d want %d",
			rr.Code, http.StatusNotFound)
	}

	if found, valid, err := files.Write("ns2", "secret2", "", ""); !found ||
		!valid || err != nil {
		t.Fatalf("Write() failed: found=%v valid=%v err=%v", found,
			valid, err)
	}
	file, ok := files.Files["ns2/secret2"]
	if !ok {
		t.Fatal("File entry not in Files map after Write()")
	}
	path := files.Path(file)
	if err := os.Remove(path); err != nil {
		t.Fatalf("os.Remove(%s): %+v", path, err)
	}
	req = httptest.NewRequest(http.MethodDelete, "/v1/pems/ns2/secret2",
		nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("DELETE /v1/pems/ns2/secret2 status: got %d want %d",
			rr.Code, http.StatusNotFound)
	}
}

func TestGetPem(t *testing.T) {
	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	files := pem.NewFiles(os.DevNull, -1, getter)

	hndlr := &pemsHndlr{
		log:       &logrus.Logger{Out: ioutil.Discard},
		files:     files,
		crtGetter: getter,
	}

	textPlain := "text/plain"
	for _, method := range []string{http.MethodGet, http.MethodHead} {
		req := httptest.NewRequest(method, "/v1/pems/ns/name", nil)
		rr := httptest.NewRecorder()
		hndlr.ServeHTTP(rr, req)
		if rr.Code != http.StatusNotImplemented {
			t.Errorf("%s /v1/pems/ns/name status: got %d want %d",
				method, rr.Code, http.StatusMethodNotAllowed)
		}
		if rr.Result().Header.Get("Content-Length") != "0" {
			t.Errorf("%s /v1/pems/ns/name Content-Length: "+
				"got %s want 0", method,
				rr.Result().Header.Get("Content-Length"))
		}
		if rr.Result().Header.Get("Content-Type") != textPlain {
			t.Errorf("%s /v1/pems/ns/name Content-Type: "+
				"got %s want %s", method,
				rr.Result().Header.Get("Content-Type"),
				textPlain)
		}
	}
}

func TestPem405(t *testing.T) {
	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	files := pem.NewFiles(os.DevNull, -1, getter)

	hndlr := &pemsHndlr{
		log:       &logrus.Logger{Out: ioutil.Discard},
		files:     files,
		crtGetter: getter,
	}

	textPlain := "text/plain"
	methods := []string{
		http.MethodPatch, http.MethodConnect, http.MethodOptions,
		http.MethodTrace, "BAZQUUX",
	}
	for _, method := range methods {
		req := httptest.NewRequest(method, "/v1/pems/ns/name", nil)
		rr := httptest.NewRecorder()
		hndlr.ServeHTTP(rr, req)
		if rr.Code != http.StatusMethodNotAllowed {
			t.Errorf("%s /v1/pems/ns/name status: got %d want %d",
				method, rr.Code, http.StatusMethodNotAllowed)
		}
		if rr.Result().Header.Get("Content-Length") != "0" {
			t.Errorf("%s /v1/pems/ns/name Content-Length: "+
				"got %s want 0", method,
				rr.Result().Header.Get("Content-Length"))
		}
		if rr.Result().Header.Get("Content-Type") != textPlain {
			t.Errorf("%s /v1/pems/ns/name Content-Type: "+
				"got %s want %s", method,
				rr.Result().Header.Get("Content-Type"),
				textPlain)
		}
	}
}

func TestAllPem(t *testing.T) {
	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	files := pem.NewFiles(os.DevNull, -1, getter)

	hndlr := &pemsHndlr{
		log:       &logrus.Logger{Out: ioutil.Discard},
		files:     files,
		crtGetter: getter,
	}

	textPlain := "text/plain"
	methods := []string{
		http.MethodGet, http.MethodHead, http.MethodPost,
		http.MethodPut, http.MethodPatch, http.MethodDelete,
		http.MethodConnect, http.MethodOptions, http.MethodTrace,
		"DEADBEEF",
	}

	for _, method := range methods {
		req := httptest.NewRequest(method, "/v1/pems", nil)
		rr := httptest.NewRecorder()
		hndlr.ServeHTTP(rr, req)
		if rr.Code != http.StatusNotImplemented {
			t.Errorf("%s /v1/pems status: got %d want %d",
				method, rr.Code, http.StatusMethodNotAllowed)
		}
		if rr.Result().Header.Get("Content-Length") != "0" {
			t.Errorf("%s /v1/pems Content-Length: got %s want 0",
				method, rr.Result().Header.
					Get("Content-Length"))
		}
		if rr.Result().Header.Get("Content-Type") != textPlain {
			t.Errorf("%s /v1/pems Content-Type: got %s want %s",
				method, rr.Result().Header.Get("Content-Type"),
				textPlain)
		}
	}
}
